using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using encinsp;

namespace cmdencinsp
{
    class Program
    {
        #region <<< Private Fields >>>

        private static String m_Directory = System.Environment.CurrentDirectory;
        private static String m_Filename = null;
        private static Boolean m_IsRecursive = false;
        private static String m_OutputFilename = null;
        private static StreamWriter m_TextWriter = null;

        #endregion <<< Private Fields >>>

        #region <<< Program Entry >>>

        static void Main(String[] args)
        {
            try
            {
                // Parse arguments
                ParseArgs(args);

                // If an output file name is specified, open it
                if (String.IsNullOrEmpty(m_OutputFilename) == false)
                {
                    Console.WriteLine("Output to file: {0}", m_OutputFilename);
                    m_TextWriter = new StreamWriter(File.Open(m_OutputFilename, FileMode.Create, FileAccess.Write, FileShare.None));
                }

                // If single file
                if (String.IsNullOrEmpty(m_Filename) == false)
                {
                    FileInfo fi = new FileInfo(m_Filename);
                    DisplayFileInfo(fi, false);
                }
                // If directory
                else
                {
                    DirectoryInfo di = new DirectoryInfo(m_Directory);
                    DisplayDirectoryInfo(di);
                }
            }
            catch (Exception ex)
            {
                // If we get an error, display it to the screen
                Console.WriteLine("Error: {0}\n{1}", ex.Message, ex.StackTrace);
            }
            finally
            {
                // Close the output file if we're writing to it
                if (m_TextWriter != null)
                {
                    m_TextWriter.Dispose();
                }
            }
        }

        #endregion <<< Program Entry >>>

        #region <<< Private Methods >>>

        /// <summary>
        /// Parse arguments
        /// </summary>
        /// <param name="args">Arguments array</param>
        private static void ParseArgs(String[] args)
        {
            // Cycle through all arguments
            for (Int32 i = 0; i < args.Length; i++)
            {
                String arg = args[i];

                // If the argument starts with a "-" or a "/"
                if (arg.StartsWith("-") || arg.StartsWith("/"))
                {
                    arg = arg.Substring(1).ToLower();

                    // -r = recursive
                    if (arg == "r")
                    {
                        m_IsRecursive = true;
                    }
                    // -o = CSV output
                    else if (arg == "o")
                    {
                        // make sure there's another argument
                        if (i == args.Length - 1)
                        {
                            Console.WriteLine("Cannot continue, no output filename specified.  Use \"encinsp ?\" for help.");
                            System.Environment.Exit(1);
                            return;
                        }

                        // get the next argument
                        i++;
                        m_OutputFilename = args[i];

                        // append ".csv" if necessary
                        if (String.Compare((new FileInfo(m_OutputFilename)).Extension, ".csv", true) != 0)
                        {
                            m_OutputFilename += ".csv";
                        }

                        // make sure the file doesn't exist
                        /*if (File.Exists(m_OutputFilename))
                        {
                            Console.WriteLine("Cannot continue, {0} already exists.  Use \"encinsp ?\" for help.", m_OutputFilename);
                            System.Environment.Exit(1);
                            return;
                        }*/

                        // make sure the file isn't a directory
                        if (Directory.Exists(m_OutputFilename))
                        {
                            Console.WriteLine("Cannot continue, {0} is a directory.  Use \"encinsp ?\" for help.", m_OutputFilename);
                            System.Environment.Exit(1);
                            return;
                        }
                    }
                    else
                    {
                        // some "-" argument we didn't recognize
                        Console.WriteLine("Invalid argument.  Use \"encinsp ?\" for help.");
                        System.Environment.Exit(1);
                        return;
                    }
                }
                else if (arg == "?")
                {
                    // display help
                    Console.WriteLine("EncInsp ver 0.1 - MP3 Encoding Inspector - Built: 2006-11-18");
                    Console.WriteLine("- Based (heavily) on Gambit's Mr QuestionMan v0.7 (http://burrrn.net)");
                    Console.WriteLine("- Converted to C# by Jud (http://cdtag.com)");
                    Console.WriteLine();
                    Console.WriteLine("usage:");
                    Console.WriteLine();
                    Console.WriteLine("[filename|directory] : scan file name or directory (default=current)");
                    Console.WriteLine("-r                   : recursive");
                    Console.WriteLine("-o [filename]        : output to csv");
                    Console.WriteLine("?                    : this screen");
                    
                    System.Environment.Exit(0);
                }
                else
                {
                    // is the argument a valid file?
                    if (File.Exists(arg))
                    {
                        m_Filename = arg;
                    }
                    else
                    {
                        // is the argument a valid directory?
                        if (Directory.Exists(arg))
                        {
                            m_Directory = arg;
                        }
                        else
                        {
                            // not a file or directory
                            Console.WriteLine("Cannot continue, {0} is not a file or directory.  Use \"encinsp ?\" for help.", arg);
                            System.Environment.Exit(1);
                            return;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Display encoder info for all files in a directory.  Recurse subdirs if necessary.
        /// </summary>
        /// <param name="baseDir">Base directory</param>
        private static void DisplayDirectoryInfo(DirectoryInfo baseDir)
        {
            // if it's recursive, go through subdirs
            if (m_IsRecursive)
            {
                foreach (DirectoryInfo di in baseDir.GetDirectories())
                {
                    DisplayDirectoryInfo(di);
                }
            }

            // cycle through all files in the directory
            foreach (FileInfo fi in baseDir.GetFiles("*.mp3"))
            {
                DisplayFileInfo(fi, m_IsRecursive);
            }
        }

        /// <summary>
        /// Display encoder info for a file
        /// </summary>
        /// <param name="fi">The file</param>
        /// <param name="displayFullName">If true, displays the full path</param>
        private static void DisplayFileInfo(FileInfo fi, bool displayFullName)
        {
            if (String.Compare(fi.Extension, ".mp3", true) != 0) return;

            DescriptiveLameTagReader reader = new DescriptiveLameTagReader(fi.FullName);

            if (m_TextWriter == null)
            {
                if (displayFullName)
                {
                    Console.WriteLine("{0} - {1} {2}", 
                        fi.FullName, 
                        reader.LameTagInfoEncoder, 
                        reader.LameTagInfoPreset);
                }
                else
                {
                    Console.WriteLine("{0} - {1} {2}", 
                        fi.Name, 
                        reader.LameTagInfoEncoder, 
                        reader.LameTagInfoPreset);
                }
            }
            else
            {
                m_TextWriter.WriteLine("\"{0}\",\"{1} {2}\"", 
                    CSVFriendly(fi.FullName), 
                    CSVFriendly(reader.LameTagInfoEncoder), 
                    CSVFriendly(reader.LameTagInfoPreset));
            }
        }

        /// <summary>
        /// Returns a CSV friendly string
        /// </summary>
        /// <param name="text">Text to fix</param>
        /// <returns>CSV friendly string</returns>
        private static String CSVFriendly(String text)
        {
            return text.Replace("\"", "\"\"");
        }

        #endregion <<< Private Methods >>>
    }
}
