using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Runtime.InteropServices;

namespace encinsp
{
    public class BasicLameTagReader
    {
        #region <<< Private Fields >>>

        private const Byte Info1Offset = 0x0D;
        private const Byte Info2Offset = 0x15;
        private const Byte Info3Offset = 0x24;
        private const Byte LAMETagOffset = 0x77;

        private LameTag m_Tag;
        private LamePreset m_PresetGuess;

        private Boolean m_IsPresetGuessNonBitrate;
        private Boolean m_IsLameTagFound;
        private UInt16 m_Preset;
        private String m_VersionString;
        private String m_VersionStringNonLameTag;

        #endregion <<< Private Fields >>>

        #region <<< Constructor >>>

        public BasicLameTagReader(String path)
        {
            // Initialize
            m_IsLameTagFound = true;

            m_Tag = new LameTag();

	        using (BinaryReader br = new BinaryReader(File.Open(path, FileMode.Open, FileAccess.Read, FileShare.Read)))
            {
                // Seek past ID3v2 tag
                Int32 startPos = ID3v2.GetTagSize(br);
                br.BaseStream.Seek(startPos, SeekOrigin.Begin);

                // Get StartOfFile structure
                StartOfFile startOfFile = StartOfFile.FromBinaryReader(br);

                // Seek past ID3v2 tag
                br.BaseStream.Seek(startPos, SeekOrigin.Begin);

                String info1 = System.Text.Encoding.ASCII.GetString(startOfFile.Info1);
                String info2 = System.Text.Encoding.ASCII.GetString(startOfFile.Info2);
                String info3 = System.Text.Encoding.ASCII.GetString(startOfFile.Info3);

		        if (info1 == "Xing" || info1 == "Info")
                {
                    br.BaseStream.Seek(Info1Offset, SeekOrigin.Current);
                }
		        else if (info2 == "Xing" || info2 == "Info")
                {
                    br.BaseStream.Seek(Info2Offset, SeekOrigin.Current);
                }
		        else if (info3 == "Xing" || info3 == "Info")
                {
                    br.BaseStream.Seek(Info3Offset, SeekOrigin.Current);
                }
		        else
                {
                    m_IsLameTagFound = true;
                }

                // Read LAME tag structure
                br.BaseStream.Seek(LAMETagOffset, SeekOrigin.Current);
                m_Tag = LameTag.FromBinaryReader(br);
                
                /// Read old LAME header
                br.BaseStream.Seek(0 - Marshal.SizeOf(typeof(LameTag)), SeekOrigin.Current);
                OldLameHeader oldLameHeader = OldLameHeader.FromBinaryReader(br);
                m_VersionStringNonLameTag = System.Text.Encoding.ASCII.GetString(oldLameHeader.VersionString);
            }

            // Set version string
            Byte[] versionString = new Byte[6];
            Int32 i;
            for (i = 0; i < 4 || (i == 4 && m_Tag.VersionString[i] == 'b'); i++) versionString[i] = m_Tag.VersionString[i];
            Array.Resize<Byte>(ref versionString, i);
            m_VersionString = System.Text.Encoding.ASCII.GetString(versionString);

            // If encoder is not LAME, set IsLameTagFound to false
            // TODO : How about other encoders that use the LAME tag?
		    if (System.Text.Encoding.ASCII.GetString(m_Tag.Encoder) != "LAME")
            {
                m_IsLameTagFound = false;
            }

            // Set preset WORD
            m_Preset = (UInt16)(((m_Tag.Surround_Preset[0] << 8) + m_Tag.Surround_Preset[1]) & 0x07FF);

            // Guess preset
            m_PresetGuess = (new PresetGuesser()).GuessPreset(
                this.VersionStringNonLameTag, /*m_Tag.VersionString*/
                m_Tag.Bitrate,
                m_Tag.Quality,
                m_Tag.TagRevision_EncodingMethod,
                m_Tag.NoiseShaping,
                m_Tag.StereoMode,
                m_Tag.EncodingFlags_ATHType,
                m_Tag.Lowpass,
                out m_IsPresetGuessNonBitrate);

        }

        #endregion <<< Constructor >>>

        #region <<< Public Properties >>>

        /// <summary>
        /// Returns the version string from the LAME tag
        /// </summary>
        public String VersionString
        {
            get
            {
                return m_VersionString;
            }
        }

        /// <summary>
        /// Returns the version string from the old LAME header (pre-3.90)
        /// </summary>
        public String VersionStringNonLameTag
        {
            get
            {
                // In versions of LAME before the LAME tag was introduced, a 16 byte version
                // string was written into the LAME header, starting at the same position as
                // the current version string.
                return m_VersionStringNonLameTag;
            }
        }

        /// <summary>
        /// Returns Encoding Method Byte
        /// </summary>
        public Byte EncodingMethod { get { return m_Tag.TagRevision_EncodingMethod; } }

        /// <summary>
        /// Returns Preset WORD
        /// </summary>
        public UInt16 Preset { get { return m_Preset; } }

        /// <summary>
        /// Returns guessed preset enum
        /// </summary>
        public LamePreset PresetGuess { get { return m_PresetGuess; } }

        /// <summary>
        /// Returns bitrate from the LAME tag (not the actual bitrate for VBR files)
        /// </summary>
        public Byte Bitrate { get { return m_Tag.Bitrate; } }

        /// <summary>
        /// Returns true if the preset is guessed to be a command-line modified version of a preset.
        /// Only applies to LAME encoded MP3's that do not have preset info stored in the LAME tag.
        /// </summary>
        public bool IsPresetGuessNonBitrate { get { return m_IsPresetGuessNonBitrate; } }

        /// <summary>
        /// Returns true if a LAME tag is present
        /// </summary>
        public bool IsLameTagFound { get { return m_IsLameTagFound; } }

        #endregion <<< Public Properties >>>
    }
}
