using System;
using System.Collections.Generic;
using System.Text;

namespace encinsp
{
    public class DescriptiveLameTagReader
    {
        #region <<< Private Fields >>>

        private BasicLameTagReader m_BasicReader;
        private UsePresetGuess m_UsePresetGuess;
        private MpegAudio m_Mpeg;
        private String m_Preset;
        private String m_PresetGuess;

        #endregion <<< Private Fields >>>

        #region <<< Constructor >>>

        public DescriptiveLameTagReader(String path)
        {
            m_Mpeg = new MpegAudio(path);

            m_BasicReader = new BasicLameTagReader(path);
            DeterminePresetRelatedValues();
        }

        #endregion <<< Constructor >>>

        #region <<< Public Properties >>>

        /// <summary>
        /// Returns true if a LAME tag is present
        /// </summary>
        public Boolean IsLameTagFound { get { return m_BasicReader.IsLameTagFound; } }

        /// <summary>
        /// Returns the version string from the LAME tag
        /// </summary>
        public String VersionString { get { return m_BasicReader.VersionString; } }

        /// <summary>
        /// Returns the version string from the old LAME header (pre-3.90)
        /// </summary>
        public String VersionStringNonLameTag { get { return m_BasicReader.VersionStringNonLameTag; } }
        
        /// <summary>
        /// Returns the preset guessing method used
        /// </summary>
        public UsePresetGuess UsePresetGuess { get { return m_UsePresetGuess; } }

        /// <summary>
        /// Returns true if the preset is guessed to be a command-line modified version of a preset.
        /// 
        /// If this value is true, the LameTagInfoPreset property will have the string "(modified)" appended, 
        /// otherwise it wil have "(guessed)" appended.
        /// 
        /// Only applies to LAME encoded MP3's that do not have preset info stored in the LAME tag.
        /// </summary>
        public Boolean IsPresetGuessNonBitrate { get { return m_BasicReader.IsPresetGuessNonBitrate; } }

        /// <summary>
        /// Returns the actual encoder preset (not guessed)
        /// </summary>
        public String Preset { get { return m_Preset; } }

        /// <summary>
        /// Returns the guessed encoder preset
        /// </summary>
        public String PresetGuess { get { return m_PresetGuess; } }

        /// <summary>
        /// Returns MPEG version and Layer
        /// </summary>
        public String LameTagInfoVersion
        {
            get
            {
                return m_Mpeg.Version + " " + m_Mpeg.Layer;
            }
        }

        /// <summary>
        /// Returns encoder
        /// </summary>
        public String LameTagInfoEncoder
        {
            get
            {
                String tmpLameTagInfoVersion;

                if (this.IsLameTagFound == false)
                {
                    tmpLameTagInfoVersion = m_Mpeg.Encoder;
                }
                else
                {
                    if (String.Compare(this.VersionString, "3.90") < 0) // FirstLameWithTag
                    {
                        // this file was encoded by an earlier version of LAME and
                        // the header only contains a version string.
                        tmpLameTagInfoVersion = "LAME " + this.VersionStringNonLameTag;
                    }
                    else
                    {
                        tmpLameTagInfoVersion = "LAME " + this.VersionString;
                    }
                }

                return tmpLameTagInfoVersion;
            }
        }

        /// <summary>
        /// Returns preset information
        /// </summary>
        public String LameTagInfoPreset
        {
            get
            {
                String tmpLameTagInfoPreset = "";

                if (this.IsLameTagFound)
                {
                    if (String.Compare(this.VersionString, "3.90") >= 0) // FirstLameWithTag
                    {
                        tmpLameTagInfoPreset = this.Preset;

                        if (this.UsePresetGuess == UsePresetGuess.UseGuess)
                        {
                            if (this.IsPresetGuessNonBitrate)
                            {
                                tmpLameTagInfoPreset = this.PresetGuess + " (modified)";
                            }
                            else
                            {
                                tmpLameTagInfoPreset = this.PresetGuess + " (guessed)";
                            }
                        }
                    }
                }

                return tmpLameTagInfoPreset;
            }
        }

        #endregion <<< Public Properties >>>

        #region <<< Private Methods >>>

        /// <summary>
        /// Determine Preset or PresetGuess values
        /// </summary>
        private void DeterminePresetRelatedValues()
        {
            m_Preset = DeterminePreset(out m_UsePresetGuess);

            if (m_UsePresetGuess == UsePresetGuess.NotNeeded)
            {
                m_PresetGuess = "";
            }
            else
            {
                m_PresetGuess = DeterminePresetGuess(ref m_UsePresetGuess);

                if (m_BasicReader.IsPresetGuessNonBitrate)
                {
                    m_PresetGuess += String.Format(" -b {0}", m_BasicReader.Bitrate);
                }
            }
        }

        /// <summary>
        /// Determine Preset
        /// </summary>
        /// <param name="usePresetGuess">Method sets this variable to the UsePresetGuess used</param>
        /// <returns>Preset string</returns>
        private String DeterminePreset(out UsePresetGuess usePresetGuess)
        {
            usePresetGuess = UsePresetGuess.NotNeeded;

            String result = "";

            Int32 preset = m_BasicReader.Preset;

            if (preset >= 8 && preset <= 320)
            {
                result = preset.ToString();
                if (m_BasicReader.EncodingMethod == 1)
                    result = "cbr " + result;
                usePresetGuess = UsePresetGuess.UseGuess;
            }
            else
            {
                switch (preset)
                {
                    case 0:
                        result = "<not stored>";
                        usePresetGuess = UsePresetGuess.UseGuess;
                        break;

                    case 410:
                        result = "V9";
                        break;

                    case 420:
                        result = "V8";
                        break;

                    case 430:
                        result = "V7";
                        break;

                    case 440:
                        result = "V6";
                        break;

                    case 450:
                        result = "V5";
                        break;

                    case 460:
                        result = "V4: preset medium";
                        break;

                    case 470:
                        result = "V3";
                        break;

                    case 480:
                        result = "V2: preset standard";
                        break;

                    case 490:
                        result = "V1";
                        break;

                    case 500:
                        result = "V0: preset extreme";
                        break;

                    case 1000:
                        result = "r3mix";
                        break;

                    case 1001:
                        result = "--alt-preset standard";
                        break;

                    case 1002:
                        result = "--alt-preset extreme";
                        break;

                    case 1003:
                        result = "--alt-preset insane";
                        break;

                    case 1004:
                        result = "--alt-preset fast standard";
                        break;

                    case 1005:
                        result = "--alt-preset fast extreme";
                        break;

                    case 1006:
                        result = "preset medium";
                        break;

                    case 1007:
                        result = "preset fast medium";
                        break;

                    case 1010:
                        result = "preset portable";	// only used in alpha versions (3.94 alpha 14 and below)
                        break;

                    case 1015:
                        result = "preset radio";	// only used in alpha versions (3.94 alpha 14 and below)
                        break;

                    default:
                        result = String.Format("<unrecognised value {0}>", preset);
                        usePresetGuess = UsePresetGuess.UseGuess;
                        break;
                }
            }

            if (m_BasicReader.EncodingMethod == 4)
            {
                if (preset == 410 || preset == 420 || preset == 430 || preset == 440 || preset == 450 ||
                    preset == 460 || preset == 470 || preset == 480 || preset == 490 || preset == 500)
                {
                    result += " (fast mode)";
                }
            }

            return result;
        }

        private String DeterminePresetGuess(ref UsePresetGuess usePresetGuess)
        {
            String Result = "";

            switch (m_BasicReader.PresetGuess)
            {
                case LamePreset.Insane:
                    Result = "--alt-preset insane";
                    break;

                case LamePreset.Extreme:
                    Result = "--alt-preset extreme";
                    break;

                case LamePreset.FastExtreme:
                    Result = "--alt-preset fast extreme";
                    break;

                case LamePreset.Standard:
                    Result = "--alt-preset standard";
                    break;

                case LamePreset.FastStandard:
                    Result = "--alt-preset fast standard";
                    break;

                case LamePreset.Medium:
                    Result = "preset medium";
                    break;

                case LamePreset.FastMedium:
                    Result = "preset fast medium";
                    break;

                case LamePreset.R3mix:
                    Result = "r3mix";
                    break;

                case LamePreset.Studio:
                    Result = "preset studio";
                    break;

                case LamePreset.CD:
                    Result = "preset cd";
                    break;

                case LamePreset.Hifi:
                    Result = "preset hifi";
                    break;

                case LamePreset.Tape:
                    Result = "preset tape";
                    break;

                case LamePreset.Radio:
                    Result = "preset radio";
                    break;

                case LamePreset.FM:
                    Result = "preset fm";
                    break;

                case LamePreset.TapeRadioFM:
                    Result = "preset tape OR preset radio OR preset fm";
                    break;

                case LamePreset.Voice:
                    Result = "preset voice";
                    break;

                case LamePreset.MWUS:
                    Result = "preset mw-us";
                    break;

                case LamePreset.MWEU:
                    Result = "preset phon+ OR preset lw OR preset mw-eu OR preset sw";
                    break;

                case LamePreset.Phone:
                    Result = "preset phone";
                    break;

                default:
                    Result = "";
                    // Only display "unable to guess" if the preset was not recorded (ie. the
                    // field value is zero.
                    if (m_BasicReader.Preset == 0)
                        usePresetGuess = UsePresetGuess.UnableToGuess;
                    else
                        usePresetGuess = UsePresetGuess.NotNeeded;
                    break;
            }

            return Result;
        }

        #endregion
    }
}
